"""
TestsProcessor Plugin
#####################

It is often required to validate network devices state against certain
criteria, test functions designed to address most common use cases in
that area.

Majority of available checks help to implement simple approach of
"if this then that" logic, for instance, if output contains this
pattern then test failed.

Dependencies:

* Nornir 3.0 and beyond
* `Cerberus module <https://pypi.org/project/Cerberus/>`_ for ``CerberusTest`` function

Test functions returns Nornir ``Result`` object and make use of these attributes:

* ``name`` - name of the test
* ``task`` - name of the task
* ``result`` -  test result ``PASS``, ``FAIL`` or ``ERROR``
* ``success`` - test success status True (PASS) or False (FAIL or ERROR)
* ``exception`` - description of failure reason
* ``test`` - test type to perform e.g. ``contains``, ``custom``, ``cerberus`` etc.
* ``criteria`` - criteria that failed the test e.g. pattern or string
* ``failed`` - this attribute not used by test functions to signal any status
  and should always be False

Running tests
=============

Running tests as simple as defining a list of dictionaries - test suite - each dictionary
represents single test definition. Reference to a particular test function API for
description of test function specific arguments it supports.

These are mandatory arguments/keys each test dictionary must contain:

* ``name`` - name of the test
* ``task`` - name of the task to check results for or list of task names to use with custom test function
* ``test`` - name of test function to run

Additional arguments/keys that test dictionary can contain:

* ``err_msg`` - string, error message to use for exception in case of test failure
* ``path`` - string, dot separated path to data to test within results
* ``report_all`` - boolean, default is False, if ``path`` evaluates to a list of items
  and ``report_all`` set to True, reports all tests, even successful ones

To simplify test functions calls, ``TestsProcessor`` implements these set of aliases
for ``test`` argument:

* ``contains`` calls ``ContainsTest``
* ``!contains`` or ``ncontains`` calls ``ContainsTest`` with  kwargs: ``{"revert": True}``
* ``contains_lines`` calls ``ContainsLinesTest``
* ``!contains_lines`` or ``ncontains_lines`` calls ``ContainsLinesTest`` with kwargs: ``{"revert": True}``
* ``contains_re`` calls ``ContainsTest`` with kwargs: ``{"use_re": True}``
* ``!contains_re`` or ``ncontains_re`` calls ``ContainsTest`` with kwargs: ``{"revert": True, "use_re": True}``
* ``equal`` calls ``EqualTest``
* ``!equal`` calls or ``nequal`` calls ``EqualTest`` with kwargs: ``{"revert": True}``
* ``cerberus`` calls ``CerberusTest``
* ``custom`` calls ``CustomFunctionTest``
* ``eval`` calls ``EvalTest``

In addition to aliases, ``test`` argument can reference actual test functions names:

* ``ContainsTest`` calls ``ContainsTest``
* ``ContainsLinesTest`` calls ``ContainsLinesTest``
* ``EqualTest`` calls ``EqualTest``
* ``CerberusTest`` calls ``CerberusTest``
* ``CustomFunctionTest`` calls ``CustomFunctionTest``
* ``EvalTest`` calls ``EvalTest``

Sample code to run tests::

    import pprint

    from nornir import InitNornir
    from nornir_salt import TestsProcessor, ResultSerializer, netmiko_send_commands

    nr = InitNornir(config_file="nornir.yaml")

    tests = [
        {
            "name": "Test NTP config",
            "task": "show run | inc ntp",
            "test": "contains",
            "pattern": "ntp server 7.7.7.8",
        },
        {
            "name": "Test Logging config",
            "task": "show run | inc logging",
            "test": "contains_lines",
            "pattern": ["logging host 1.1.1.1", "logging host 1.1.1.2"]
        },
        {
            "name": "Test BGP peers state",
            "task": "show bgp ipv4 un summary",
            "test": "!contains_lines",
            "pattern": ["Idle", "Active", "Connect"]
        },
        {
            "task": "show run | inc ntp",
            "name": "Test NTP config",
            "expr": "assert '7.7.7.8' in result, 'NTP server 7.7.7.8 not in config'",
            "test": "eval",
        }
    ]

    nr_with_tests = nr.with_processors([
        TestsProcessor(tests, remove_tasks=True)
    ])

    # netmiko_send_commands maps commands to sub-task names
    results = nr_with_tests.run(
        task=netmiko_send_commands,
        commands=[
            "show run | inc ntp",
            "show run | inc logging",
            "show bgp ipv4 un summary"
        ]
    )

    results_dictionary = ResultSerializer(results, to_dict=False, add_details=False)

    pprint.pprint(results_dictionary)

    # should print something like:
    #
    # [{'host': 'IOL1', 'name': 'Test NTP config', 'result': 'PASS'},
    # {'host': 'IOL1', 'name': 'Test Logging config', 'result': 'PASS'},
    # {'host': 'IOL1', 'name': 'Test BGP peers state', 'result': 'FAIL'},
    # {'host': 'IOL2', 'name': 'Test NTP config', 'result': 'PASS'},
    # {'host': 'IOL2', 'name': 'Test Logging config', 'result': 'PASS'},
    # {'host': 'IOL2', 'name': 'Test BGP peers state', 'result': 'PASS'}]

Notes on ``path`` attribute. ``path`` attribute allows to run tests against portions
of overall results, but works only if results are structured data, e.g. nested dictionary
or list of dictionaries. For example::

    import pprint
    from nornir import InitNornir
    from nornir_salt import TestsProcessor, ResultSerializer, nr_test

    nr = InitNornir(config_file="nornir.yaml")

    tests = [
        {
            "test": "eval",
            "task": "show run interface",
            "name": "Test MTU config",
            "path": "interfaces.*",
            "expr": "assert result['mtu'] > 9000, '{} MTU less then 9000'.format(result['interface'])"
        }
    ]

    nr_with_tests = nr.with_processors([
        TestsProcessor(tests, remove_tasks=True)
    ])

    # nr_test function echoes back ret_data_per_host as task results
    output = nr_with_tests.run(
        task=nr_test,
        ret_data_per_host={
            "IOL1": {
                "interfaces":  [
                    {"interface": "Gi1", "mtu": 1500},
                    {"interface": "Gi2", "mtu": 9200},
                ]
            },
            "IOL2": {
                "interfaces":  [
                    {"interface": "Eth1/9", "mtu": 9600}
                ]
            }
        },
        name="show run interface"
    )

    check_result = ResultSerializer(output, add_details=True, to_dict=False)
    # pprint.pprint(check_result)
    # [{'changed': False,
    #   'criteria': '',
    #   'diff': '',
    #   'exception': 'Gi1 MTU less then 9000',
    #   'failed': True,
    #   'host': 'IOL1',
    #   'name': 'Test MTU config',
    #   'result': 'FAIL',
    #   'success': False,
    #   'task': 'show run interface',
    #   'test': 'eval'},
    #  {'changed': False,
    #   'criteria': '',
    #   'diff': '',
    #   'exception': None,
    #   'failed': False,
    #   'host': 'IOL2',
    #   'name': 'Test MTU config',
    #   'result': 'PASS',
    #   'success': True,
    #   'task': 'show run interface',
    #   'test': 'eval'}]

In above example path ``interfaces.*`` tells ``TestsProcessor`` to retrieve data from
results under ``interfaces`` key, single star ``*`` symbol tells to iterate over list
items, instead of star, list item index can be given as well, e.g. ``interfaces.0``.

Tests Reference
===============

.. autoclass:: nornir_salt.plugins.processors.TestsProcessor.TestsProcessor
.. autofunction:: nornir_salt.plugins.processors.TestsProcessor.ContainsTest
.. autofunction:: nornir_salt.plugins.processors.TestsProcessor.ContainsLinesTest
.. autofunction:: nornir_salt.plugins.processors.TestsProcessor.EqualTest
.. autofunction:: nornir_salt.plugins.processors.TestsProcessor.CerberusTest
.. autofunction:: nornir_salt.plugins.processors.TestsProcessor.EvalTest
.. autofunction:: nornir_salt.plugins.processors.TestsProcessor.CustomFunctionTest
"""
import logging
import re
import traceback

from nornir.core.inventory import Host
from nornir.core.task import AggregatedResult, MultiResult, Result, Task

log = logging.getLogger(__name__)

try:
    from cerberus import Validator

    HAS_CERBERUS = True
except ImportError:
    log.debug("Failed to import Cerberus library, install: pip install cerberus")
    HAS_CERBERUS = False

# return datum template dictionary
test_result_template = {
    "name": "",  # name of the test
    "task": "",  # name of the task
    "result": "PASS",  # test result PASS, FAIL, ERROR
    "success": True,  # success status
    "failed": False,  # failure status
    "exception": None,  # description of failure reason
    "test": None,  # test type to perform .g. contains
    "criteria": "",  # criteria that failed the test e.g. pattern or string
}


def _get_result_by_path(data, path, host):
    """
    Helper generator function to iterate over data and yield
    ``nornir.core.task.Result`` object with result attribute set to data
    at given path.

    :param data: (dict or list) structured data to retrieve data subset from
    :param path: (list) list of path items to get from data
    :param host: (obj)  Nornir host object
        :return result: ``nornir.core.task.Result`` object with data at given path
    """
    if path == []:
        yield Result(host=host, result=data)
    elif isinstance(data, dict):
        for result in _get_result_by_path(data[path[0]], path[1:], host):
            yield result
    elif path[0].isdigit() and isinstance(data, list):
        for result in _get_result_by_path(data[int(path[0])], path[1:], host):
            yield result
    elif path[0] == "*" and isinstance(data, list):
        for item in data:
            for result in _get_result_by_path(item, path[1:], host):
                yield result


def EvalTest(host, result, expr, revert=False, err_msg=None, globs={}, **kwargs):
    """
    Function to check result running python built-in ``Eval`` or ``Exec``
    function against provided python expression.

    This function in its use cases sits in between pre-built test function such as
    ``ContainsTest`` or ``EqualTest`` and running custom Python test function using
    ``CustomFunctionTest`` function. ``Eval`` allows to use any python expressions
    that evaluates to True or False without the need to write custom functions.

    If expression string starts with ``assert``, will use ``Exec`` function,
    uses ``Eval`` for everything else.

    Eval and Exec functions' ``globals`` dictionary populated with ``result`` and
    ``host`` variables, ``result`` contains ``nornir.core.task.Result`` result attribute
    while ``host`` references Nornir host object. This allows to use expressions
    like this::

        "'7.7.7.7' in result"
        "assert '7.7.7.8' in result, 'NTP server 7.7.7.8 not in config'"
        "len(result.splitlines()) == 3"

    Eval and Exec functions' ``globals`` dictionary attribute merged with ``**globs``
    supplied to ``EvalTest`` function call, that allows to use any additional variables
    or functions. For example, below is equivalent to running ``contains_lines`` test::

        tests = [
            {
                "test": "eval",
                "task": "show run | inc logging",
                "name": "Test Syslog config",
                "expr": "all(map(lambda line: line in result, lines))",
                "globs": {
                    "lines": ["logging host 1.1.1.1", "logging host 2.2.2.2"]
                },
                "err_msg": "Syslog config is wrong"
            }
        ]

    ``lines`` variable shared with ``eval`` globals space, allowing to reference it
    as part of expression.

    .. warning: using eval and exec could be dangerous, running tests from untrusted
        sources generally a bad idea as any custom code can be executed on the system.

    :param host: (obj) Nornir host object
    :param result: (obj) ``nornir.core.task.Result`` object
    :param expr: (str) Python expression to evaluate
    :param revert: (bool) if True, changes results to opposite - check for inequality
    :param err_msg: (str) exception message to use on test failure
    :param globs: (dict) dictionary to use as ``eval/exec`` ``globals`` space
    :param kwargs: (dict) any additional ``**kwargs`` keyword arguments to include in return Result object
    :return result: ``nornir.core.task.Result`` object with test results
    """
    ret = test_result_template.copy()
    ret.update(kwargs)

    try:
        if expr.strip().startswith("assert"):
            check_result = exec(
                expr, {"result": result.result, "host": host, **globs}, {}
            )
        else:
            check_result = eval(
                expr, {"result": result.result, "host": host, **globs}, {}
            )
        if check_result is False:
            ret.update({"result": "FAIL", "success": False})
            ret["exception"] = err_msg if err_msg else "Expression evaluated to False"
    except AssertionError as e:
        ret.update({"result": "FAIL", "success": False})
        ret["exception"] = err_msg if err_msg else (str(e) or "AssertionError")
    except:
        ret.update({"result": "ERROR", "success": False})
        ret["exception"] = err_msg if err_msg else traceback.format_exc()

    # revert results
    if revert:
        if ret["success"] is False:
            ret.update({"result": "PASS", "success": True})
            ret["exception"] = None
        elif ret["success"] is True:
            ret.update({"result": "FAIL", "success": False})
            ret["exception"] = err_msg if err_msg else "Pattern and output equal"
    return Result(host=host, **ret)


def _cerberus_validate_item(validator_engine, data, schema, ret_data):
    """Helper function to avoid code repetition for Cerberus validation"""
    res = validator_engine.validate(document=data, schema=schema)
    if not res:
        ret_data.update(
            {"result": "FAIL", "success": False, "exception": validator_engine.errors}
        )
    try:
        ret_data["name"] = ret_data["name"].format(**data)
    except KeyError:
        pass
    return ret_data


def CerberusTest(host, result, schema, allow_unknown=True, **kwargs):
    """
    Function to check results using ``Cerberus`` module schema. Results must be a structured
    data - dictionary, list - strings and other types of data not supported.

    :param host: (obj) Nornir host object
    :param result: ``nornir.core.task.Result`` object
    :param schema: (dictionary) Cerberus schema definition to us for validation
    :param allow_uncknown: (bool) Cerberus allow unknown parameter, default is True
    :param kwargs: (dict) any additional ``**kwargs`` keyword arguments to include in return Result object

    .. warning:: Cerberus library only supports validation of dictionary structures,
        while nested elements could be lists, as a result, ``CerberusTest`` function
        was coded to support validation of dictionary or list of dictionaries results.

    .. note:: ``kwargs`` ``name`` key value formatted using python format function supplying
        dictionary being validated as arguments
    """
    # form ret structure
    ret = test_result_template.copy()
    ret.update(kwargs)

    # run check
    if not HAS_CERBERUS:
        ret.update({"result": "ERROR", "success": False})
        ret[
            "exception"
        ] = "Failed to import Cerberus library, install: pip install cerberus"
        return Result(host=host, **ret)
    validator_engine = Validator()
    validator_engine.allow_unknown = allow_unknown

    # validate dictionary results
    if isinstance(result.result, dict):
        ret = _cerberus_validate_item(validator_engine, result.result, schema, ret)
        return Result(host=host, **ret)
    # validate list of dictionaries results
    elif isinstance(result.result, list):
        validation_results = []
        for item in result.result:
            if not isinstance(item, dict):
                continue
            ret_copy = ret.copy()
            ret_copy = _cerberus_validate_item(validator_engine, item, schema, ret_copy)
            validation_results.append(Result(host=host, **ret_copy))
        return validation_results
    else:
        raise TypeError(
            "nornir-salt:CerberusTest unsupported results type '{}', supported - dictionary, list".format(
                type(result.result)
            )
        )


def _load_custom_fun_from_text(function_text, function_name, globals_dictionary={}):
    """
    Helper function to load custom function code from text using
    Python ``exec`` built-in function
    """
    assert (
        function_name in function_text
    ), "nornir-salt:CustomFunctionTest no '{}' function in function text".format(
        function_name
    )

    data = {}
    glob_dict = {
        "__builtins__": __builtins__,
        "False": False,
        "True": True,
        "None": None,
    }
    glob_dict.update(globals_dictionary)

    # load function by running exec
    exec(compile(function_text, "<string>", "exec"), glob_dict, data)

    # add extracted functions to globals for recursion to work
    glob_dict.update(data)

    return data[function_name]


def CustomFunctionTest(
    host,
    result,
    function_file=None,
    function_text=None,
    function_call=None,
    function_name="run",
    function_kwargs={},
    use_all_tasks=False,
    globals_dictionary={},
    **kwargs
):
    """
    Wrapper around calling custom function to perform results checks.

    :param host: (obj) Nornir host object
    :param result: ``nornir.core.task.Result`` object
    :param function_name: (str) function name, default is ``run``
    :param function_file: (str) OS path to file with ``function_name`` function
    :param function_text: (str) Python code text for ``function_name`` function
    :param function_call: (callable) reference to callable python function
    :param use_all_tasks: (bool) if True passes all host results to custom function, default False
    :param globals_dictionary: (dict) dictionary to merge with global space of the custom function,
      used only if ``function_file`` or ``function_text`` arguments provided.
    :param function_kwargs: (dict) ``**function_kwargs`` to pass on to custom function
    :param kwargs: (dict) any additional key word arguments to include in results

    .. warning:: ``function_file`` and ``function_text`` use ``exec`` function
       to compile python code, using test functions from untrusted sources can
       be dangerous.

    Custom functions should accept one positional argument for results following these rules:

    * if ``task`` is a string result is ``nornir.core.task.Result``
    * if ``task`` is a list of task names result is a list of ``nornir.core.task.Result`` objects
      of corresponding tasks
    * if ``use_all_tasks`` set to True result is ``nornir.core.task.MultiResult`` object

    Any additional parameters can be passed to custom test function using ``function_kwargs``
    arguments.

    Custom function can return a dictionary or a list of dictionaries to include in
    results. Each dictionary can have any keys, but it is recommended to have at least
    these keys:

    * ``exception`` - error description if any
    * ``result`` - "PASS", "FAIL" or "ERROR" string
    * ``success`` - boolean True or False

    If a list returned by custom function, each list item forms individual result item.

    If custom test function returns empty list, empty dictionary or None or True test considered
    successful and dictionary added to overall results with ``result`` key set to ``PASS``.

    If custom test function returns False test outcome considered unsuccessful and dictionary
    added to overall results with ``result`` key set to ``FAIL``.

    Sample custom test function to accept ``Result`` object when ``use_all_tasks`` set to False and
    ``task`` is a sting representing name of the task::

        def custom_test_function(result):
            # result is nornir.core.task.Result object
            if "7.7.7.8" not in result.result:
                return {
                    "exception": "Server 7.7.7.8 not in config",
                    "result": "FAIL",
                    "success": False
                }

    Sample custom test function to accept ``MultiResult`` object when ``use_all_tasks`` set to True::

        def custom_test_function(result):
            # result is nornir.core.task.MultiResult object - list of nornir.core.task.Result objects
            ret = []
            for item in result:
                if item.result == None: # skip empty results
                    continue
                elif item.name == "show run | inc ntp":
                    if "7.7.7.8" not in item.result:
                        ret.append({
                            "exception": "NTP Server 7.7.7.8 not in config",
                            "result": "FAIL",
                            "success": False
                        })
                elif item.name == "show run | inc logging":
                    if "1.1.1.1" not in item.result:
                        ret.append({
                            "exception": "Logging Server 1.1.1.1 not in config",
                            "result": "FAIL",
                            "success": False
                        })
            return ret
    """
    # form ret structure
    ret = test_result_template.copy()
    ret.update(kwargs)

    # load and compile custom function
    try:
        if function_text:
            test_function = _load_custom_fun_from_text(
                function_text, function_name, globals_dictionary
            )
        elif function_file:
            with open(function_file, encoding="utf-8") as f:
                test_function = _load_custom_fun_from_text(
                    f.read(), function_name, globals_dictionary
                )
        elif function_call:
            test_function = function_call
        else:
            raise RuntimeError(
                "nornir-salt:CustomFunctionTest no custom function found."
            )
    except:
        msg = "nornir-salt:CustomFunctionTest function loading error:\n{}".format(
            traceback.format_exc()
        )
        log.error(msg)
        ret.update({"result": "ERROR", "success": False, "exception": msg})
        return Result(host=host, **ret)
    # run custom function
    try:
        test_function_result = test_function(result, **function_kwargs)
    except:
        msg = "nornir-salt:CustomFunctionTest function run error:\n{}".format(
            traceback.format_exc()
        )
        log.error(msg)
        ret.update({"result": "ERROR", "success": False, "exception": msg})
        return Result(host=host, **ret)
    # form and return results
    if (
        test_function_result == []
        or test_function_result == {}
        or test_function_result is None
        or test_function_result is True
    ):
        return Result(host=host, **ret)
    elif test_function_result is False:
        ret.update({"result": "FAIL", "success": False})
        return Result(host=host, **ret)
    elif isinstance(test_function_result, list):
        ret_list = []
        for item in test_function_result:
            ret_copy = ret.copy()
            ret_copy.update(item)
            ret_list.append(Result(host=host, **ret_copy))
        return ret_list
    elif isinstance(test_function_result, dict):
        ret.update(test_function_result)
        return Result(host=host, **ret)
    else:
        raise TypeError(
            "nornir-salt:CustomFunctionTest test function returned unsupported results type: {}".format(
                type(test_function_result)
            )
        )


def EqualTest(host, result, pattern, revert=False, err_msg=None, **kwargs):
    """
    Function to check result is equal to the pattern.

    :param host: (obj) Nornir host object
    :param result: (obj) ``nornir.core.task.Result`` object
    :param pattern: (any) string, dict, list or any other object to check for equality
    :param revert: (bool) if True, changes results to opposite - check for inequality
    :param err_msg: (str) exception message to use on test failure
    :param kwargs: (dict) any additional ``**kwargs`` keyword arguments to include in return Result object
    :return result: ``nornir.core.task.Result`` object with test results
    """
    # form ret structure
    ret = test_result_template.copy()
    ret.update(kwargs)
    if isinstance(pattern, str):
        ret["criteria"] = (
            pattern.replace("\n", "\\n")
            if len(pattern) < 25
            else pattern[0:24].replace("\n", "\\n")
        )

    # run the check
    if pattern != result.result:
        ret.update({"result": "FAIL", "success": False})
        ret["exception"] = err_msg if err_msg else "Pattern and output not equal"
    # revert results
    if revert:
        if ret["success"] is False:
            ret.update({"result": "PASS", "success": True})
            ret["exception"] = None
        elif ret["success"] is True:
            ret.update({"result": "FAIL", "success": False})
            ret["exception"] = err_msg if err_msg else "Pattern and output equal"
    return Result(host=host, **ret)


def ContainsLinesTest(
    host,
    result,
    pattern,
    use_re=False,
    count=None,
    revert=False,
    err_msg=None,
    **kwargs
):
    """
    Function to check that all lines contained in result output.

    Tests each line one by one, this is the key difference compared to
    ``ContainsTest`` function, where whole pattern checked for presence in
    output from device.

    :param host: (obj) Nornir host object
    :param result: (obj) ``nornir.core.task.Result`` object
    :param pattern: (str or list) multiline string or list of lines to check
    :param use_re: (bool) if True uses re.search to check for line pattern in output
    :param count: (int) check exact number of line pattern occurrences in the output
    :param revert: (bool) if True, changes results to opposite - check lack of lines in output
    :param err_msg: (str) exception message to use on test failure
    :param kwargs: (dict) any additional ``**kwargs`` keyword arguments to include in return Result object
    :return result: ``nornir.core.task.Result`` object with test results
    """
    # form ret structure
    ret = test_result_template.copy()
    ret.update(kwargs)

    # run the check
    lines_list = pattern.splitlines() if isinstance(pattern, str) else pattern
    for line in lines_list:
        check_result = ContainsTest(
            host=host,
            result=result,
            pattern=line,
            use_re=use_re,
            count=count,
            revert=revert,
            err_msg=err_msg,
        )
        if not check_result.success:
            ret.update({"result": "FAIL", "success": False})
            ret["exception"] = check_result.exception
            ret["criteria"] = check_result.criteria
            break
    return Result(host=host, **ret)


def ContainsTest(
    host,
    result,
    pattern,
    use_re=False,
    count=None,
    revert=False,
    err_msg=None,
    **kwargs
):
    """
    Function to check if pattern contained in output of given result.

    :param host: (obj) Nornir host object
    :param result: (obj) ``nornir.core.task.Result`` object
    :param pattern: (str) pattern to check containment for
    :param use_re: (bool) if True uses re.search to check for pattern in output
    :param count: (int) check exact number of pattern occurrences in the output
    :param revert: (bool) if True, changes results to opposite - check lack of containment
    :param err_msg: (str) exception message to use on test failure
    :param kwargs: (dict) any additional ``**kwargs`` keyword arguments to include in return Result object
    :return result: ``nornir.core.task.Result`` object with test results
    """
    # form ret structure
    ret = test_result_template.copy()
    ret.update(kwargs)
    ret["criteria"] = (
        pattern.replace("\n", "\\n")
        if len(pattern) < 25
        else pattern[0:24].replace("\n", "\\n")
    )

    # add count to return results
    if count:
        ret["count"] = count
    # run the check
    if use_re:
        if not re.search(pattern, result.result):
            ret.update({"result": "FAIL", "success": False})
            ret["exception"] = err_msg if err_msg else "Regex pattern not in output"
    elif count and result.result.count(pattern) != count:
        ret.update({"result": "FAIL", "success": False})
        ret["exception"] = (
            err_msg if err_msg else "Pattern not in output {} times".format(count)
        )
    elif pattern not in result.result:
        ret.update({"result": "FAIL", "success": False})
        ret["exception"] = err_msg if err_msg else "Pattern not in output"
    # revert results if requested to do so
    if revert:
        if ret["success"] is False:
            ret.update({"result": "PASS", "success": True})
            ret["exception"] = None
        elif ret["success"] is True:
            ret.update({"result": "FAIL", "success": False})
            if use_re:
                ret["exception"] = err_msg if err_msg else "Regex pattern in output"
            elif count:
                ret["exception"] = (
                    err_msg if err_msg else "Pattern in output {} times".format(count)
                )
            else:
                ret["exception"] = err_msg if err_msg else "Pattern in output"
    return Result(host=host, **ret)


test_functions_dispatcher = {
    "contains": {"fun": ContainsTest, "kwargs": {}},
    "ncontains": {"fun": ContainsTest, "kwargs": {"revert": True}},
    "!contains": {"fun": ContainsTest, "kwargs": {"revert": True}},
    "contains_re": {"fun": ContainsTest, "kwargs": {"use_re": True}},
    "!contains_re": {"fun": ContainsTest, "kwargs": {"revert": True, "use_re": True}},
    "ncontains_re": {"fun": ContainsTest, "kwargs": {"revert": True, "use_re": True}},
    "contains_lines": {"fun": ContainsLinesTest, "kwargs": {}},
    "!contains_lines": {"fun": ContainsLinesTest, "kwargs": {"revert": True}},
    "ncontains_lines": {"fun": ContainsLinesTest, "kwargs": {"revert": True}},
    "contains_lines_re": {"fun": ContainsLinesTest, "kwargs": {"use_re": True}},
    "!contains_lines_re": {
        "fun": ContainsLinesTest,
        "kwargs": {"revert": True, "use_re": True},
    },
    "ncontains_lines_re": {
        "fun": ContainsLinesTest,
        "kwargs": {"revert": True, "use_re": True},
    },
    "equal": {"fun": EqualTest, "kwargs": {}},
    "!equal": {"fun": EqualTest, "kwargs": {"revert": True}},
    "nequal": {"fun": EqualTest, "kwargs": {"revert": True}},
    "cerberus": {"fun": CerberusTest, "kwargs": {}},
    "custom": {"fun": CustomFunctionTest, "kwargs": {}},
    "eval": {"fun": EvalTest, "kwargs": {}},
}


class TestsProcessor:

    """
    TestsProcessor designed to run a series of tests for Nornir
    tasks results.

    :param tests: (list of dictionaries) list of tests to run
    :param remove_tasks: (bool) if True (default) removes tasks output from results
    :param kwargs: (any) if provided, ``**kwargs`` will form a single test item
    :param failed_only: (bool) if True, includes only failed tests in results, default is False
    """

    def __init__(self, tests=None, remove_tasks=True, failed_only=False, **kwargs):
        self.tests = tests if tests else [kwargs]
        self.remove_tasks = remove_tasks
        self.len_tasks = None
        self.failed_only = failed_only

    def task_started(self, task: Task) -> None:
        pass

    def task_instance_started(self, task: Task, host: Host) -> None:
        pass

    def task_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        """
        Method to iterate over individual hosts's result after task/sub-tasks completion.
        """
        try:
            # record the len of tasks to clean them up if required
            if self.remove_tasks:
                self.len_tasks = len(result)
            # do the tests
            for test in self.tests:
                test = test.copy()

                # if test item is a list, transform it to dictionary
                if isinstance(test, list):
                    test = {
                        "task": test[0],
                        "test": test[1],
                        "pattern": test[2],
                        "name": test[3] if len(test) == 4 else None,
                    }
                    if test["test"] in ["eval", "EvalTest"]:
                        test["expr"] = test.pop("pattern")

                # make sure we have test name defined
                if not test.get("name"):
                    test["name"] = "{} {} {}..".format(
                        test["task"], test["test"], test.get("pattern", "")[:9]
                    )

                # get task results to use; use all results
                if test.get("use_all_tasks") is True:
                    test["result"] = result
                # use subset of task results
                elif isinstance(test["task"], list):
                    test["result"] = []
                    for task_result in result:
                        if task_result.name in test["task"]:
                            test["result"].append(task_result)
                # use results for single task only
                else:
                    # try to find task by matching it's name
                    for task_result in result:
                        if task_result.name == test["task"]:
                            test["result"] = task_result
                            break
                    else:
                        # use first task if only one test and one task given
                        tasks = [t for t in result if not hasattr(t, "skip_results")]
                        if len(self.tests) == 1 and len(tasks) == 1:
                            test["result"] = tasks[0]
                        else:
                            log.warning(
                                "nornir-salt:TestsProcessor: no results for task '{}'".format(
                                    test["task"]
                                )
                            )
                            continue

                # get test function and function kwargs
                if test["test"] in test_functions_dispatcher:
                    test_func = test_functions_dispatcher[test["test"]]["fun"]
                    test.update(test_functions_dispatcher[test["test"]]["kwargs"])
                elif test["test"] in globals() and "Test" in test["test"]:
                    test_func = globals()[test["test"]]
                else:
                    raise NameError(
                        "nornir-salt:TestsProcessor unsupported test function '{}'".format(
                            test["test"]
                        )
                    )

                # run the test
                try:
                    # run test for data at given path
                    if test.get("path"):
                        report_all = test.pop("report_all", False)
                        res = [
                            test_func(host=host, result=item, **test)
                            for item in _get_result_by_path(
                                data=test.pop("result").result,
                                path=test.pop("path").split("."),
                                host=host,
                            )
                        ]
                        # leave only failed results
                        if not report_all:
                            res = [i for i in res if i.success is False]
                            # add single successful test if no tests failed
                            if not res:
                                ret = test_result_template.copy()
                                ret.update(test)
                                _ = ret.pop("expr", None)
                                res = Result(host=host, **ret)
                    else:
                        res = test_func(host=host, **test)
                except:
                    msg = "nornir-salt:TestsProcessor run error:\n{}".format(
                        traceback.format_exc()
                    )
                    log.error(msg)
                    ret = test_result_template.copy()
                    ret.update(test)
                    ret.update({"result": "ERROR", "success": False, "exception": msg})
                    res = Result(host=host, **ret)

                if isinstance(res, list):
                    result.extend(res)
                else:
                    result.append(res)
        except:
            result.append(
                Result(
                    host=host,
                    exception=traceback.format_exc(),
                    result=traceback.format_exc(),
                    success=False,
                    name="nornir-salt:TestsProcessor task_instance_completed error",
                )
            )

    def subtask_instance_started(self, task: Task, host: Host) -> None:
        pass

    def subtask_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        pass

    def task_completed(self, task: Task, result: AggregatedResult) -> None:

        # remove tasks with device's output
        if self.remove_tasks:
            for hostname, results in result.items():
                if len(results) >= self.len_tasks:
                    for i in range(0, self.len_tasks):
                        _ = results.pop(0)

        # remove non failed tasks if requested to do so
        if self.failed_only:
            for hostname, results in result.items():
                good_tests = []
                for index, i in enumerate(results):
                    if hasattr(i, "success") and i.success is True:
                        good_tests.append(index)
                # pop starting from last index to preserve lower indexes
                for i in reversed(good_tests):
                    _ = results.pop(i)



"""
pygnmi_call
###########

Task plugin to manage devices over gNMI protocol. 

Requires `PyGNMI library <https://pypi.org/project/pygnmi/>`_ to be installed::

    pip install pygnmi

This task plugin is a wrapper around ``gNMIclient`` connection object
and allows to execute any of its methods supplying method name using
``call`` attribute.

Sample code to run ``pygnmi_call`` task::

    from nornir import InitNornir
    from nornir_salt import pygnmi_call

    nr = InitNornir(config_file="config.yaml")
 
    # get device capabilities
    capabilities = nr.run(
        task=pygnmi_call,
        call="capabilities"
    )
    
    # get interfaces configuration
    get_output = nr.run(
        task=pygnmi_call,
        call="get",
        path=["openconfig-interfaces:interfaces"]
    )

    # update interface description
    update_output = nr.run(
        task=pygnmi_call,
        call="update",
        update=[
            (
                "openconfig-interfaces:interfaces/interface[name=Loopback100]/config", 
                {"description": "Done by gNMI"}
            )
        ]
    )

    # delete interface configuration
    delete_output = nr.run(
        task=pygnmi_call,
        call="delete",
        delete=[
            "openconfig-interfaces:interfaces/interface[name=Loopback1234]"
        ]
    )

    # replace interface configuration
    replace_output = nr.run(
        task=pygnmi_call,
        call="replace",
        replace=[
            (
                "openconfig-interfaces:interfaces/interface[name=Loopback1234]/config", 
                {"name": "Loopback1234", "description": "New"}
            )
        ]
    )
    
In addition to calling ``gNMIclient`` methods, extra ``call`` functions supported
such as ``help``, ``dir``, ``delete``, ``replace`` and ``update``. Extra functions
can be invoked in the same way as ``gNMIclient`` connection object methods by passing 
their name as a ``call`` attribute::

    from nornir import InitNornir
    from nornir_salt import pygnmi_call

    nr = InitNornir(config_file="config.yaml")
    
    available_methods = nr.run(
        task=pygnmi_call,
        call="dir",
    )

pygnmi_call Reference
=====================

.. autofunction:: nornir_salt.plugins.tasks.pygnmi_call.pygnmi_call

Additional Call Functions 
=========================

delete
------
.. autofunction:: nornir_salt.plugins.tasks.pygnmi_call._call_delete

dir
---
.. autofunction:: nornir_salt.plugins.tasks.pygnmi_call._call_dir

help
----
.. autofunction:: nornir_salt.plugins.tasks.pygnmi_call._call_help

replace
-------
.. autofunction:: nornir_salt.plugins.tasks.pygnmi_call._call_replace

update
------
.. autofunction:: nornir_salt.plugins.tasks.pygnmi_call._call_update
"""
import traceback
import logging

from nornir.core.task import Result, Task
from nornir_salt.plugins.connections.PyGNMIPlugin import CONNECTION_NAME

log = logging.getLogger(__name__)


def _call_dir(connection, **kwargs):
    """Function to return a list of ``gNMIclient`` available methods/operations"""
    methods = (list(dir(connection))) + ["dir", "help", "update", "delete", "replace"]
    result = sorted(
        [m for m in set(methods) if (not m.startswith("_") and not m.isupper())]
    )
    return result


def _call_help(connection, method_name: str, **kwargs):
    """
    Helper function to return docstring for requested method

    :param method_name: (str) name of method or function to return docstring for
    """
    if "_call_{}".format(method_name) in globals():
        function_obj = globals()["_call_{}".format(method_name)]
    else:
        function_obj = getattr(connection, method_name)
    h = function_obj.__doc__ if hasattr(function_obj, "__doc__") else ""
    return h


def _call_update(connection, path: list, **kwargs):
    """
    Udate function helps to update configuration for elements matched by
    single path string.

    This function effectively takes arguments provided and uses ``gNMIclient``
    ``set`` method supplying it with ``update`` argument list that consists of
    single tuple element, this tuple's first item is a provided path string with
    second item being a dictionary of provided ``**kwargs`` containing configuration
    to be updated.

    :param connection: (obj) ``gNMIclient`` object
    :param path: (list) list with single item - path to element to update config for
    :param kwargs: (dict) configuration parameters to update

    Sample code to run ``replace`` function task::

        from nornir import InitNornir
        from nornir_salt import pygnmi_call

        nr = InitNornir(config_file="config.yaml")

        output = nr.run(
            task=pygnmi_call,
            call="update",
            path="openconfig-interfaces:interfaces/interface[name=Loopback100]/config",
            description="Updated Loopback Description"
        )
    """
    return connection.set(update=[(path[0], dict(kwargs))])


def _call_replace(connection, path: list, **kwargs):
    """
    Replace function helps to replace configuration for elements matched by
    single path string.

    This function effectively takes arguments provided and uses ``gNMIclient``
    ``set`` method supplying it with ``replace`` argument list that consists of
    single tuple element, this tuple's first item is a provided path string with
    second item being a dictionary of provided ``**kwargs`` containing configuration
    to be replaced.

    :param connection: (obj) ``gNMIclient`` object
    :param path: (list) list with single item - path to element to update config for
    :param kwargs: (dict) configuration parameters to replace

    Sample code to run ``replace`` function task::

        from nornir import InitNornir
        from nornir_salt import pygnmi_call

        nr = InitNornir(config_file="config.yaml")

        output = nr.run(
            task=pygnmi_call,
            call="replace",
            path="openconfig-interfaces:interfaces/interface[name=Loopback100]/config",
            name="Loopback100",
            description="Loopback Description"
        )
    """
    return connection.set(replace=[(path[0], dict(kwargs))])


def _call_delete(connection, path: list, **kwargs):
    """
    Delete function helps to delete configuration elements matched by provided paths
    strings.

    This function effectively takes arguments provided and uses ``gNMIclient``
    ``set`` method supplying it with ``delete`` argument list that consists of
    path items strings.

    :param connection: (obj) ``gNMIclient`` object
    :param path: (list) path items to delete

    Sample code to run ``delete`` function task::

        from nornir import InitNornir
        from nornir_salt import pygnmi_call

        nr = InitNornir(config_file="config.yaml")

        output = nr.run(
            task=pygnmi_call,
            call="delete",
            path=["openconfig-interfaces:interfaces/interface[name=Loopback100]"]
        )
    """
    return connection.set(delete=path)


def pygnmi_call(task: Task, call: str, name_arg: str = None, **kwargs) -> Result:
    """
    Task to call one of PyGNMI ``gNMIclient`` object methods or one of
    additional helper functions.

    :param call: (str) ``gNMIclient`` connection object method to call
    :param arg: (list) any ``*args`` to use with call method
    :param kwargs: (dict) any ``**kwargs`` to use with call method
    :param name_arg: (str) used as "name" argument with call method, need it
        only because "name" argument used by "Nornir.run" method itself ans collides
        with the case when need to pass gNMI path ``name`` argument to this task

    Special handling given to ``path``, ``delete``, ``replace`` and
    ``updated`` kwargs arguments to comply with ``gNMIclient`` requirements:

    1. If ``path`` is a string, convert it to a list splitting it by ``,`` character
    2. If ``delete`` is a string, convert it to a list splitting it by ``,`` character
    3. If ``replace`` is a list, transform each list item to a tuple
    4. If ``update`` is a list, transform each list item to a tuple
    """
    # update task name
    task.name = call

    # check if "_name_arg" in kwargs, use it as "name", this happens
    # if actual "name" argument given on nr.gnmi call by saltstack
    if name_arg:
        kwargs["name"] = name_arg

    # transform arguments for set delete and get calls from string to list
    if isinstance(kwargs.get("path"), str):
        kwargs["path"] = [i.strip() for i in kwargs["path"].split(",") if i.strip()]
    if isinstance(kwargs.get("delete"), str):
        kwargs["delete"] = [i.strip() for i in kwargs["delete"].split(",") if i.strip()]
    # convert set replace and set update list items to tuples
    if isinstance(kwargs.get("replace"), list):
        kwargs["replace"] = [
            tuple(i) for i in kwargs["replace"] if isinstance(i, (list, tuple))
        ]
    if isinstance(kwargs.get("update"), list):
        kwargs["update"] = [
            tuple(i) for i in kwargs["update"] if isinstance(i, (list, tuple))
        ]

    # get gNMIclient connection object
    connection = task.host.get_connection(CONNECTION_NAME, task.nornir.config)

    log.debug("nornir_salt:pygnmi_call call '{}', kwargs: '{}'".format(call, kwargs))

    # check if need to call one of helper function
    if "_call_{}".format(call) in globals():
        result = globals()["_call_{}".format(call)](connection, **kwargs)
    # call connection object method otherwise
    else:
        result = getattr(connection, call)(**kwargs)

    return Result(host=task.host, result=result)



"""
netmiko_send_command_ps
#######################

Send command string to device using promptless (ps) approach. Can be used for any 
command, including commands that change device prompt. Multiple commands can be sent
separated by '\n' newline.

.. image:: ../_images/promptless_mode_v0.1.png

Promptless mode allows to detect end of output from device without relying on timers or
correct prompt matching (hence the name - promptless). This mode still uses pattern to 
decide if device finished emitting data, but that pattern is not dependent on device's prompt
regex matching.

Each reading cycle, data from device read as fast as possible until device either finishes or 
pose to prepare more data. Latter case detected and handled using read timeout timer and checking 
if new data received. To detect when device finishes producing output, algorithm sends two space
character to device and checks in next read cycle if last line contains two additional spaces,
concluding that end of output detected if so and continue cycling otherwise.

Overall its similar to how Humans interact with device prompt to verify that its still operational 
- try hitting space or type something in terminal to see if its appears on the screen.

Dependencies:

* `nornir-netmiko module <https://pypi.org/project/nornir-netmiko/>`_ required

netmiko_send_command_ps sample usage
=====================================

Code to invoke ``netmiko_send_command_ps`` task::

    from nornir import InitNornir
    from nornir_salt import netmiko_send_command_ps

    nr = InitNornir(config_file="config.yaml")
    
    commands = '''
    show ip int brief
    conf t
    interface loopback 100
    ip address 1.1.1.100 255.255.255.255
    end
    show ip int brief
    wr
    '''
    
    output = nr.run(
        task=netmiko_send_command_ps,
        commands="show run",
        netmiko_kwargs={}
    )

    output_multiline = nr.run(
        task=netmiko_send_command_ps,
        commands=commands,
        netmiko_kwargs={}
    )
    
netmiko_send_command_ps returns
================================

Returns Nornir results object with individual tasks names set
equal to commands sent to device.

netmiko_send_command_ps reference
==================================

.. autofunction:: nornir_salt.plugins.tasks.netmiko_send_command_ps.netmiko_send_command_ps
.. autofunction:: nornir_salt.plugins.tasks.netmiko_send_command_ps.send_command_ps
"""

import time
import logging
from typing import Any
from nornir.core.task import Result, Task
from difflib import get_close_matches

try:
    from nornir_netmiko.connections import CONNECTION_NAME

    HAS_NETMIKO = True
except ImportError:
    HAS_NETMIKO = False

log = logging.getLogger(__name__)


def send_command_ps(
    self,
    command_string,
    read_timeout=30,
    timeout=120,
    inter_loop_sleep=0.1,
    initial_sleep=0.1,
    strip_prompt=True,
    strip_command=True,
    normalize=True,
    cutoff=0.6,
    nowait=False,
):
    """
    Execute command_string_ps on the SSH channel using promptless (ps) approach. Can be used
    for any commands, including commands that change prompt. Multiple commands can be sent
    separated by '\\n' newline.

    :param command_string: (str) The command(s) to be executed on the remote device.
    :param read_timeout: (int) Timeout in seconds to wait for data from devices, default 30s, if
        set to -1 will wait indefinitely
    :param timeout: (int) Absolute timeout in seconds of overall wait, default 120s, if
        set to -1 will wait indefinitely
    :param inter_loop_sleep: (int) Interval in seconds to sleep between reading loops, default 0.1s
    :param initial_sleep: (int) time to sleep after sending command, default 0.1s
    :param strip_prompt: (bool) Remove the trailing router prompt from the output (default: True).
    :param strip_command: (bool) Remove the echo of the command from the output (default: True).
    :param normalize: (bool) Ensure the proper enter is sent at end of command (default: True).
    :param nowait: (bool) Default is False, if True sends command and returns immediately without
        waiting for prompt right after ``initial_sleep`` timer elapsed.
    """
    data_received = ""
    previous_last_line = ""
    no_data_elapsed = 0
    start_time = time.time()

    # add a list of previously matched last lines
    if not hasattr(self, "prompts_seen"):
        self.prompts_seen = set([self.find_prompt().strip()])
        time.sleep(
            3
        )  # need to wait for device to emit find_prompt() output so that self.clear_buffer() can catch it
        log.debug(
            "send_command_ps formed prompts_seen set: {}".format(self.prompts_seen)
        )

    if normalize:
        command_string = self.normalize_cmd(command_string)

    self.clear_buffer()
    self.write_channel(command_string)

    # initial sleep
    time.sleep(initial_sleep)

    # check if need to return immediately after sending commands
    if nowait:
        return "DONE"

    # Main loop to get output from device
    while True:
        if read_timeout != -1 and no_data_elapsed > read_timeout:
            raise TimeoutError(
                "send_command_ps {}s read_timeout expired".format(read_timeout)
            )
        if timeout != -1 and (time.time() - start_time) > timeout:
            raise TimeoutError("send_command_ps {}s timeout expired".format(timeout))

        time.sleep(inter_loop_sleep)
        no_data_elapsed += inter_loop_sleep

        # read data from channel
        chunk = self._read_channel()
        if chunk:
            no_data_elapsed = 0
            data_received += chunk
            new_last_line = data_received.splitlines()[-1]
        else:
            continue

        # junos returns \x07 instead of space, replace it with spaces
        if "\x07" in new_last_line:
            new_last_line = new_last_line.replace("\x07", " ")

        # check if new_last_line looks similar to any of the previous device prompts
        matched = get_close_matches(
            new_last_line.strip(), list(self.prompts_seen), 1, cutoff
        )
        log.debug(
            "send_command_ps: last line '{}' similar prompt - '{}'".format(
                new_last_line.strip(), matched
            )
        )
        if matched:
            # Detect end of output by sending two space chars and checking if received it back in next cycle
            is_end = "{}  ".format(previous_last_line) == new_last_line
            log.debug(
                "send_command_ps: EOF detection - new_last_line: '{}'; previous_last_line: {}; is_end: {}".format(
                    [new_last_line], [previous_last_line], is_end
                )
            )
            if is_end:
                self.prompts_seen.add(new_last_line.strip())
                break
            previous_last_line = new_last_line
            self.write_channel("  ")
    output = self._sanitize_output(
        data_received,
        strip_command=strip_command,
        command_string=command_string,
        strip_prompt=strip_prompt,
    )
    return output


def netmiko_send_command_ps(
    task: Task, command_string: str, enable: bool = False, **kwargs: Any
) -> Result:
    """
    Patch netmiko connection object with send_command_ps method and
    execute it.

    :param command_string: (str) Command to execute on the remote network device.
    :param enable: (bool) Set to True to force Netmiko .enable() call.
    :param kwargs: (dict) Additional arguments to pass to send_command_ps method.
    :return result object: with result of the show command
    """
    net_connect = task.host.get_connection(CONNECTION_NAME, task.nornir.config)
    if enable:
        net_connect.enable()

    if "send_command_ps" not in dir(net_connect):
        setattr(net_connect, "send_command_ps", send_command_ps)

    result = net_connect.send_command_ps(net_connect, command_string, **kwargs)
    return Result(host=task.host, result=result)



"""
ConnectionsPool
###############

`PyGNMI library <https://pypi.org/project/pygnmi/>`_ connection plugin to interact 
with devices over `gNMI <https://github.com/openconfig/reference/blob/master/rpc/gnmi/gnmi-specification.md>`_ 
protocol.

This plugin maintains long running gNMI connection to devices, if this behavior 
not desirable, consider using Nornir host's ``close_connection`` method to close
gNMI connection.

ConnectionsPool reference
=========================

.. autofunction:: nornir_salt.plugins.connections.ConnectionsPool.ConnectionsPool
"""
import logging
import time
from typing import Any, Dict, Optional
from nornir.core.configuration import Config
from multiprocessing import Lock

log = logging.getLogger(__name__)


class ConnObjWrap:
    """
    Helper class to wrap connection object into
    to implement context manager
    """

    def __init__(self, conection, host):
        self.connection = conection
        self.host = host
        self.lock = Lock()

    def __enter__(self):
        self.lock.acquire(block=False)
        return self.connection

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.lock.release()
        log.debug(
            "nornir-salt:ConnectionPool {} connection released '{}'".format(
                self.host.name, self.connection
            )
        )


class ConnectionsPool:
    """

    Sample inventory::

        host-1:
          hostname: 192.168.16.20
          username: admin
          password: admin
          port: 22
          connection_options:
            netmiko:
              platform: ariste_eos
            ConnectionsPool:
              extras:
                max: 3
    """  # noqa

    def __init__(self):
        self.connections = []
        self.parameters = {}
        self.max = 3
        self.connection = None

    def open(
        self,
        hostname: Optional[str],
        username: Optional[str],
        password: Optional[str],
        port: Optional[int],
        platform: Optional[str],
        extras: Optional[Dict[str, Any]] = None,
        configuration: Optional[Config] = None,
    ) -> None:

        self.parameters = {
            "hostname": hostname,
            "username": username,
            "password": password,
            "port": port,
            "platform": platform,
            "extras": extras,
            "configuration": configuration,
        }

        self.max = extras.get("max", 3)

        self.connection = self

    def close(self) -> None:
        for c in self.connections():
            c.connection.close()

    def __enter__(self, connection_name, host):
        host_conn = self.get_connection(connection_name, host)
        with host_conn() as c:
            return c

    def __exit__(self, exc_type, exc_value, exc_traceback):
        pass

    def get_connection(self, connection_name, host, timeout=3):
        """
        Function to get connection and lock it.
        """
        host_conn = None
        stime = time.time()

        while int(time.time() - stime) < timeout:

            # see if has unused connections
            for c in self.connections:
                if c.lock.acquire(block=False):
                    host_conn = c
                    break
            if host_conn:
                break

            # establish additional connection
            if len(self.connections) < self.max:
                conn = host.get_connection(
                    connection_name,
                    configuration=self.parameters["configuration"],
                )
                c = ConnObjWrap(conn, host)
                _ = host.connections.pop(connection_name)
                c.lock.acquire()
                self.connections.append(c)
                host_conn = c
                if len(self.connections) == self.max:
                    host.connections[connection_name] = self
                break

            time.sleep(0.1)

        # if reached this point it means failed to get connection
        else:
            raise RuntimeError(
                "nornir-salt:ConnectionPool {} failed to acquire connection '{}'".format(
                    host.name, connection_name
                )
            )

        log.debug(
            "nornir-salt:ConnectionPool {} acquired connection '{}'".format(
                host.name, host_conn.connection
            )
        )

        return host_conn



import setuptools

__version__ = "0.7.0"
__author__ = "Denis Mulyalin"

with open("README.md", "r") as f:
    README = f.read()

setuptools.setup(
    name="nornir_salt",
    version=__version__,
    author=__author__,
    author_email="d.mulyalin@gmail.com",
    description="Nornir plugins used with SALTSTACK",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/dmulyalin/nornir-salt",
    packages=setuptools.find_packages(),
    extras_require={},
    classifiers=[
        "Topic :: Utilities",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
    ],
    python_requires=">=3.6",
    install_requires=[
        "nornir==3.*",
    ],
    entry_points="""
    [nornir.plugins.inventory]
    DictInventory=nornir_salt:DictInventory

    [nornir.plugins.runners]
    QueueRunner=nornir_salt:QueueRunner
    RetryRunner=nornir_salt:RetryRunner

    [nornir.plugins.connections]
    ncclient=nornir_salt:NcclientPlugin
    http=nornir_salt:HTTPPlugin
    pygnmi=nornir_salt:PyGNMIPlugin
    ConnectionsPool=nornir_salt:ConnectionsPool
    """,
)



"""
PyGNMIPlugin
############

`PyGNMI library <https://pypi.org/project/pygnmi/>`_ connection plugin to interact 
with devices over `gNMI <https://github.com/openconfig/reference/blob/master/rpc/gnmi/gnmi-specification.md>`_ 
protocol.

This plugin maintains long running gNMI connection to devices, if this behavior 
not desirable, consider using Nornir host's ``close_connection`` method to close
gNMI connection.

PyGNMIPlugin reference
========================

.. autofunction:: nornir_salt.plugins.connections.PyGNMIPlugin.PyGNMIPlugin
"""
from typing import Any, Dict, Optional
from nornir.core.configuration import Config

try:
    from pygnmi.client import gNMIclient

    HAS_PYGNMI = True
except ImportError:
    HAS_PYGNMI = False

CONNECTION_NAME = "pygnmi"


class PyGNMIPlugin:
    """
    Full list of inventory extras see `PyGNMI Docs <https://github.com/akarneliuk/pygnmi>`_
    for ``gNMIclient`` class.

    Sample inventory::

        host-1:
          hostname: 192.168.16.20
          username: admin
          password: admin
          port: 2022
          connection_options:
            pygnmi:
              extras:
                insecure: True
                gnmi_timeout: 10

    Anything under inventory ``extras`` section passed on to PyGNMI
    ``gNMIclient`` class object instantiation.
    """  # noqa

    def open(
        self,
        hostname: Optional[str],
        username: Optional[str],
        password: Optional[str],
        port: Optional[int],
        platform: Optional[str],
        extras: Optional[Dict[str, Any]] = None,
        configuration: Optional[Config] = None,
    ) -> None:
        extras = extras or {}

        parameters: Dict[str, Any] = {
            "target": (hostname, port or 6030),
            "username": username,
            "password": password,
        }

        parameters.update(extras)

        self.client = gNMIclient(**parameters)
        self.connection = self.client.connect()

    def close(self) -> None:
        self.client.close()
        self.connection.close()



from .ToFileProcessor import ToFileProcessor
from .TestsProcessor import TestsProcessor
from .DiffProcessor import DiffProcessor
from .DataProcessor import DataProcessor
from .SaltEventProcessor import SaltEventProcessor

__all__ = (
    "ToFileProcessor",
    "TestsProcessor",
    "DiffProcessor",
    "DataProcessor",
    "SaltEventProcessor",
)
